import time
from collections import defaultdict
from dataclasses import dataclass, field
from homonculi.config import Configuration
from homonculi.quizzes.loader import QuizFile
from homonculi.schemas.prompt_function import PromptCollection


@dataclass
class ResultRow:
    question_id: str
    student_id: str
    attempt: int
    answer: str
    confidence: str
    score: float
    correct: bool
    points: float
    weighted: float
    system_error: str
    feedback: str
    mistake: str
    reason: str

    def to_csv(self, with_confidence: bool = True) -> list:
        return (
            [self.question_id, self.student_id, self.attempt, self.answer]
            + ([self.confidence] if with_confidence else [])
            + [
                self.score,
                self.correct,
                self.points,
                self.weighted,
                self.system_error,
                self.feedback,
                self.mistake,
                self.reason,
            ]
        )


RESULT_HEADERS = [
    "QuestionID",
    "StudentID",
    "Attempt",
    "Answer",
    "Confidence",
    "Score",
    "Correct",
    "PointsPossible",
    "Weighted",
    "SystemError",
    "Feedback",
    "Mistake",
    "Reason",
]
WITHOUT_CONFIDENCE = [header for header in RESULT_HEADERS if header != "Confidence"]


@dataclass
class JobData:
    config: Configuration
    quiz: QuizFile
    prompts: PromptCollection
    runs: list = field(default_factory=list)
    time_started: int = field(default_factory=lambda: int(time.time()))
    time_finished: int | None = None
    created_prompts: dict[str, dict[str, str]] = field(
        default_factory=lambda: defaultdict(dict)
    )

    def to_json(self) -> str:
        return {
            "config": self.config.dumps_json(),
            "quiz": self.quiz.to_json(),
            "prompts": self.prompts.to_json(),
            "runs": self.runs,
            "time_started": self.time_started,
            "time_finished": self.time_finished,
        }



import shlex
from dataclasses import dataclass
import simple_parsing
from simple_parsing.helpers import Serializable


@dataclass
class Configuration(Serializable):

    quiz_path: str
    """The path to the quiz file. If it's a txt file, then it should be a list of relative paths to actual questions. Yaml/JSON file can either be a list of relative paths, or the actual questions (in the documented Quiz structure)."""

    quiz_feedback_path: str | None = None
    """If given, this is assumed to be the autograding logic of the JSON, separate from the quiz questions stored in quiz_path."""

    attempts: int = 3
    """The number of times to ask the LLM to solve each quiz question. """

    quiz_load_mode: str = "guess"
    """The mode to use to load the quiz file. Can be 'guess', 'flat', 'single', 'split', or 'spreadsheet'."""

    quiz_load_format: str = "guess"
    """The format to use to load the quiz file. Can be 'guess', 'json', 'yaml', 'txt', or 'csv'."""

    output_format: str = "both"
    """The format to use to output the results. Can be 'csv', 'log', or 'both'."""

    output_dir: str = "results"
    """The directory to save the results to. """

    results_path: str = "{quiz}_{time}.csv"
    """The path template to save the results to. Can use {quiz} and {time} as placeholders."""

    log_path: str = "{quiz}_{time}.yml"
    """The path template to save the log to. Can use {quiz} and {time} as placeholders."""

    seed: int = 42
    """The seed to use for the random number generator (not on GPT, just locally). """

    backup_save_threshold: int = 60 * 2
    """The number of seconds between each backup save; useful for long runs. """

    ask_for_feedback_on_answer: bool = True
    """Whether to ask the LLM for feedback on the feedback. """

    include_confidence: bool = True
    """Whether to include the confidence in the output. """

    student_prompt_dirs: str = "./prompts/students/"
    """A semicolon-separated list of directories containing student prompt files. """

    student_prompts: str = "./prompts/students/default.yaml"
    """A comma-separated list of student prompt files to use; matches the filename if an extension is present, otherwise uses the internal name in the prompt file (searching through the student_prompt_dirs). """

    question_format_prompt_path: str = "./prompts/questions/default.yaml"
    """The path to the question format prompt file; this will be used to construct the actual question function. """

    feedback_format_prompt_path: str = "./prompts/feedback/default.yaml"
    """The path to the feedback format prompt file; this will be used to construct the feedback on grading function. """

    gpt_api_key_path: str = "./openai_key.txt"
    """The path to the file containing the API key for OpenAI's API. """

    gpt_api_key: str | None = None
    """The API key for OpenAI's API. """

    gpt_model: str = "gpt-4o-mini"
    """The model to use for the GPT API. """

    gpt_temperature: float = 0.5
    """The temperature to use for the GPT API; controls some of the randomness. """

    gpt_top_p: float = 0.9
    """ The top_p value to use for the GPT API; controls the diversity of the output. """

    gpt_max_tokens: int = 1000
    """The maximum number of tokens to generate for the GPT API. """


def parse_configuration(args=None):
    parser = simple_parsing.ArgumentParser(add_config_path_arg=True)

    parser.add_arguments(Configuration, dest="config")

    if isinstance(args, str):
        args = shlex.split(args)

    return parser.parse_args(args)



"""

A few different ways to use this:

```shell
python simulate.py --quiz_path
```

Quizzes can be provided:
- As a single JSON file
- As a split JSON file with questions and autograding code
- As a collection of YAML files including an index.yaml
- As a remote canvas identifier using Waltz

The simulator will:
- Run in immediate mode
- Run in batch mode

The output will be:
- A CSV file with the results
- A detailed log file with the prompts and runs

By default, the system will send two queries per question:
- First, it will ask GPT to answer the question
- Second, it will show GPT its feedback, and ask for a response

You can override any part of the prompting:
- The student prompt
- The question format prompt
- The feedback format prompt

You can also override the GPT model and parameters.
"""

from pprint import pprint
import sys
import time
import random
import copy
import csv
from dataclasses import dataclass
import yaml
from tqdm import tqdm
from tqdm.contrib import itertools
from collections import defaultdict
import json
from pathlib import Path

from homonculi.config import Configuration, parse_configuration
from homonculi.file_utilities import add_to_csv, backup_log, start_csv
from homonculi.job import RESULT_HEADERS, JobData, ResultRow
from homonculi.quizzes.questions import QuizQuestion, post_process_answer
from homonculi.quizzes.quizzes import check_quiz_question
from homonculi.quizzes.loader import load_file
from homonculi.openai_api import OpenAI
import homonculi.quizzes as quizzes
from homonculi.schemas.prompt_function import (
    PromptFunction,
    load_prompts,
    make_prompt_for_feedback,
)
from homonculi.schemas.prompt_function import make_prompt_for_question_type
import logging

logger = logging.getLogger("homunculi")


def simulate_students(config: Configuration):
    """
    Simulate student working on a quiz
    """
    logger.info("Starting simulation")
    quiz = load_file(config)
    prompts = load_prompts(config)
    job = JobData(config, quiz, prompts)
    grader = OpenAI(config)

    store_results = config.output_format in ["both", "csv"]
    store_logs = config.output_format in ["both", "log"]
    if store_results:
        result_path = config.results_path.format(quiz=quiz.name, time=job.time_started)
        result_path = Path(config.output_dir) / result_path
        start_csv(result_path, RESULT_HEADERS)
    if store_logs:
        log_path = config.log_path.format(quiz=quiz.name, time=job.time_started)
        log_path = Path(config.output_dir) / log_path
        backup_log(job, log_path)
    last_dumped_log = time.time()

    estimated_total = (
        len(quiz.questions) * len(prompts.student_prompts) * config.attempts
    )
    progress_bar = tqdm(
        itertools.product(
            quiz.questions, prompts.student_prompts, range(config.attempts)
        ),
        total=estimated_total,
    )  # smooth=0.1

    if config.seed:
        random.seed(config.seed)

    for question, student_prompt, attempt_id in progress_bar:
        student_id = student_prompt["metadata"]["name"]
        progress_bar.set_description(f"{question.id}/{student_id}/{attempt_id}")
        created_student_prompt = make_prompt_for_question_type(
            question,
            student_prompt,
            prompts.question_format_prompt,
            config.include_confidence,
        )
        # Record the prompt if it is new
        if question.id not in job.created_prompts:
            job.created_prompts[question.id] = {}
        if student_id not in job.created_prompts[question.id]:
            job.created_prompts[student_id] = created_student_prompt.to_json()

        # Make the actual chat completion and submit to GPT
        # TODO: Allow the user to provide context (e.g., previous question, group question)
        conversation = [
            {
                "role": "system",
                "content": [
                    {"type": "text", "text": created_student_prompt.body},
                ],
            }
        ]
        potential_images = question.get_images()
        if potential_images:
            for image in potential_images:
                conversation[0]["content"].append(
                    {"type": "image_file", "image_url": image}
                )
        result, full_result = grader.run_function(
            conversation, created_student_prompt.function
        )
        answer = post_process_answer(
            question, result["answer"], prompts.question_format_prompt
        )
        if answer.system_error:
            logger.error(answer.system_error)
            continue

        # Check the answer
        score, correct, feedback = check_quiz_question(
            question, question["check"], answer.answer
        )

        row = ResultRow(
            question.id,
            student_id,
            attempt_id,
            answer.answer if answer.system_error is None else answer.original_answer,
            result.get("confidence", ""),
            score,
            correct,
            question.points,
            score * question.points,
            answer.system_error or "",
            feedback,
            "",
            "",
        )
        # Handle checking the question
        if config.ask_for_feedback_on_answer:
            created_feedback_prompt = make_prompt_for_feedback(
                correct, prompts.feedback_format_prompt
            )
            conversation.append(
                {"role": "system", "content": created_feedback_prompt.body}
            )
            feedback, full_feedback = grader.run_function(
                conversation, created_feedback_prompt.function
            )
            row.mistake = feedback["mistake"]
            row.reason = feedback["reason"]

        if store_results:
            add_to_csv(result_path, row.to_csv(config.include_confidence))

        if store_logs:
            dumped_model = full_result.model_dump()
            dumped_model["prompt"] = copy.deepcopy(conversation)
            job.runs.append(dumped_model)

        time_passed = time.time() - last_dumped_log

        if time_passed > config.backup_save_threshold:
            logger.info(f"Time passed: {time_passed} seconds. Saving log.")
            progress_bar.set_description("Saving log...")

            backup_log(job, log_path)
            last_dumped_log = time.time()

    job.time_finished = int(time.time())

    if store_logs:
        backup_log(job, log_path)


def main():
    config = parse_configuration(sys.argv[1:])
    simulate_students(config.config)


if __name__ == "__main__":
    main()



def compare_string_equality(submitted, expected) -> bool:
    if not submitted and bool(expected):
        return False
    submitted = str(submitted).strip()
    if isinstance(expected, (str, float, int)):
        return str(submitted) == str(expected)
    else:
        # Assume list of possible answers
        return submitted in expected



import json
import io
import time
import openai
import base64
import requests
from dataclasses import dataclass
import logging
from homonculi.config import Configuration

logger = logging.getLogger("homunculi")


class OpenAI:
    """
    Simple wrapper around OpenAI's API to make it easier to call functions, send prompts, generate images, and
    create batch files. This class is intended to be used as a singleton, and should be instantiated once and
    passed around as needed.

    The class also handles things like retries and rate limiting, and will automatically retry failed requests.

    Args:
        config: The configuration object to use for the OpenAI API key and other settings.
    """

    def __init__(self, config: Configuration):
        self.config = config
        if config.gpt_api_key:
            gpt_api_key = config.gpt_api_key
        elif config.gpt_api_key_path:
            with open(config.gpt_api_key_path) as f:
                gpt_api_key = f.read().strip()
        else:
            raise ValueError(
                "OpenAI API key not provided via CLI, file, or environment variable!"
            )
        self.client = openai.OpenAI(api_key=gpt_api_key)
        if not self.client.api_key:
            raise ValueError("OpenAI API key not set!")

    def encode_image(self, image_file):
        return base64.b64encode(image_file).decode("utf-8")

    def get_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.config.gpt_api_key}",
        }

    def create_completion(self, messages, function):
        return self.client.chat.completions.create(
            model=self.config.gpt_model,
            messages=self._reformat_image_messages(messages),
            tools=[{"type": "function", "function": function}],
            tool_choice={"type": "function", "function": {"name": function["name"]}},
            # functions=[function],
            # function_call={'name': function['name']},
            # TODO: Enable these options!
            # temperature=self.config.gpt_temperature,
            # top_p=self.config.gpt_top_p,
            # max_tokens=self.config.gpt_max_tokens,
            # response_format={"type": "json_object"}
        )

    def run_function(self, messages, function, attempts=3, retry_delay=10):
        def is_valid(response):
            return not response.choices or not response.choices[0].message.tool_calls

        execution = lambda: self.create_completion(messages, function)
        response = self._run_gpt(execution, is_valid, function, attempts, retry_delay)
        tool_calls = response.choices[0].message.tool_calls[0]

        try:
            args = json.loads(tool_calls.function.arguments)
        except json.JSONDecodeError as e:
            logger.error(f"Invalid JSON returned from OpenAI API: {e}")
            time.sleep(retry_delay)
            return self._run_gpt(execution, is_valid, function, attempts - 1)

        for expected_arg in function["parameters"]["required"]:
            if expected_arg not in args:
                logger.error(
                    f"Missing required parameter in OpenAI response: {expected_arg}"
                )
                time.sleep(retry_delay)
                return self._run_gpt(execution, is_valid, function, attempts - 1)

        return args, response

    def _run_gpt(self, execution, is_valid, function, attempts=3, retry_delay=10):
        """
        Runs a prompt through OpenAI's api which calls a function, and parses the result.

        Args:
            messages: A list of messages to pass to the OpenAI api call
            function: The function to pass to the OpenAI api call
        """
        if attempts <= 0:
            # TODO: Allow attempt count to be configurable
            logger.error("Too many attempts to run prompt, aborting.")
            raise Exception("Too many attempts to run prompt, aborting.")
        try:
            response = execution()
            do_retry = False
        except openai.RateLimitError as e:
            # Handle rate limit error (we recommend using exponential backoff)
            logger.error(f"OpenAI API request exceeded rate limit: {e}")
            retry_delay *= 2
            do_retry = True
        except openai.APIConnectionError as e:
            # Handle connection error here
            logger.error(f"Failed to connect to OpenAI API: {e}")
            raise Exception(f"Failed to connect to OpenAI API: {e}")
        except openai.APIStatusError as e:
            # Handle API error here, e.g. retry or log
            logger.error(
                f"OpenAI API returned an API Error: {e.status_code}\n{e.response}\n{e.response.content}"
            )
            do_retry = True

        if do_retry:
            logger.error(
                f"Retrying prompt, waiting {retry_delay} seconds ({attempts} attempts left)"
            )
            time.sleep(retry_delay)
            return self._run_gpt(execution, is_valid, function, attempts - 1)

        if is_valid(response):
            logger.error(f"Response returned from OpenAI API was not valid: {response}")
            time.sleep(retry_delay)
            return self._run_gpt(execution, is_valid, function, attempts - 1)

        return response

    def _reformat_image_messages(self, messages):
        for message in messages:
            for content in message["content"]:
                if not isinstance(content, str) and content["type"] == "image_file":
                    content["type"] = "image_url"
                    base64_image = self.encode_image(content["image_url"])
                    content["image_url"] = {
                        "url": f"data:image/jpeg;base64,{base64_image}"
                    }
        return messages

    def prepare_image_messages(self, prompt, image_data):
        base64_image = self.encode_image(image_data)
        messages = [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"},
                    },
                ],
            }
        ]
        return messages

    def run_image_prompt(self, prompt, image_data, attempts=3, retry_delay=10):
        """
        Runs a prompt through OpenAI's api which calls a function, and parses the result.

        Args:
            image_data: The image data to pass to the Open AI api call
        """
        if attempts <= 0:
            logger.error("Too many attempts to run prompt, aborting.")
            raise Exception("Too many attempts to run prompt, aborting.")

        def execution():
            return self.client.chat.completions.create(
                model=self.config.gpt_model,
                messages=self.prepare_image_messages(prompt, image_data),
                max_tokens=self.config.gpt_max_tokens,
            )

        def is_valid(response):
            return not response.choices or not response.choices[0].message.content

        response = self._run_gpt(execution, is_valid, attempts, retry_delay)

        return response

    def generate_image(
        self, prompt, size="1024x1024", quality="standard", model="dall-e-3"
    ):
        try:
            response = self.client.images.generate(
                prompt=prompt, size=size, quality=quality
            )
        except openai.OpenAIError as e:
            logger.error(f"Error generating image: {e}")
            raise Exception(f"Error generating image: {e}")

        if not response.data:
            logger.error(f"Invalid response from OpenAI API: {response}")
            raise Exception(f"Invalid response from OpenAI API: {response}")

        return response.data[0].url, response

    def upload_string_as_file(self, contents):
        io_file = io.BytesIO(contents.encode())
        file_object = self.client.files.create(file=io_file, purpose="batch")
        return file_object

    def create_batch_file(self, prompt_set, function):
        """
        Creates a batch file using the OpenAI API.

        Args:
            prompts: A list of prompts to pass to the OpenAI api call
            function: The function to pass to the OpenAI api call
        """
        messages = [self.create_completion(prompt, function) for prompt in prompt_set]
        batch_file = io.StringIO(json.dumps(messages))
        batch_input_file = self.client.files.create(file=batch_file, purpose="batch")

    def submit_batch_file(self, batch_input_file_id, description):
        batch = self.client.batches.create(
            input_file_id=batch_input_file_id,
            endpoint="/v1/chat/completions",
            completion_window="24h",
            metadata={
                "description": description,
            },
        )
        if batch.errors:
            raise Exception(f"Error submitting batch file: {batch.errors}")
        return batch

    def check_batch_file(self, batch_id):
        return self.client.batches.retrieve(batch_id)

    def get_batch_file(self, batch_id):
        batch = self.client.batches.retrieve(batch_id)
        output_file_id = batch.output_file_id
        if not output_file_id:
            raise Exception("Batch file is not ready for retrieval yet.")
        return self.client.files.content(output_file_id).text



"""
TODO: Find a real package to do this with, don't roll your own!!
"""

import random
import math


# NOTE: In the future (3.10), can use built-in statistics.correlation
# from statistics import correlation
def correlation(x, y):
    # Assume len(x) == len(y)
    n = len(x)
    sum_x = float(sum(x))
    sum_y = float(sum(y))
    sum_x_sq = sum(xi * xi for xi in x)
    sum_y_sq = sum(yi * yi for yi in y)
    psum = sum(xi * yi for xi, yi in zip(x, y))
    num = psum - (sum_x * sum_y / n)
    den = pow((sum_x_sq - pow(sum_x, 2) / n) * (sum_y_sq - pow(sum_y, 2) / n), 0.5)
    if den == 0:
        return 0
    return num / den


def binconf(p, N, c=0.95):
    """
    Calculate binomial confidence interval based on the number of positive and
    negative events observed.

    Parameters
    ----------
    p: int
      number of positive events observed
    N: int
      number of total events observed
    c : optional, [0,1]
      confidence percentage. e.g. 0.95 means 95% confident the probability of
      success lies between the 2 returned values

    Returns
    -------
    theta_low  : float
      lower bound on confidence interval
    theta_high : float
      upper bound on confidence interval
    """
    p, N = float(p), float(N)

    if N == 0.0:
        return (0.0, 1.0)

    p = p / N
    z = normcdfi(1 - 0.5 * (1 - c))

    a1 = 1.0 / (1.0 + z * z / N)
    a2 = p + z * z / (2 * N)
    a3 = z * math.sqrt(p * (1 - p) / N + z * z / (4 * N * N))

    return (a1 * (a2 - a3), a1 * (a2 + a3))


def erfi(x):
    """Approximation to inverse error function"""
    a = 0.147  # MAGIC!!!
    a1 = math.log(1 - x * x)
    a2 = 2.0 / (math.pi * a) + a1 / 2.0

    return sign(x) * math.sqrt(math.sqrt(a2 * a2 - a1 / a) - a2)


def sign(x):
    if x < 0:
        return -1
    if x == 0:
        return 0
    if x > 0:
        return 1


def normcdfi(p, mu=0.0, sigma2=1.0):
    """Inverse CDF of normal distribution"""
    if mu == 0.0 and sigma2 == 1.0:
        return math.sqrt(2) * erfi(2 * p - 1)
    else:
        return mu + math.sqrt(sigma2) * normcdfi(p)


def safe_sample(group, n):
    if len(group) < n:
        return group
    return random.sample(group, n)



import yaml
import csv
import json
from pathlib import Path
import os
import frontmatter


def start_csv(output_path: str, header):
    """
    Start a csv file with a header
    """
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with open(output_path, "w", newline="", encoding="utf-8") as f:
        csv_writer = csv.writer(f)
        csv_writer.writerow(header)


def add_to_csv(output_path: str, row):
    """
    Add feedback to a csv file
    """
    with open(output_path, "a", newline="", encoding="utf-8") as f:
        csv_writer = csv.writer(f)
        csv_writer.writerow(row)


def backup_log(info, meta_path: str):
    os.makedirs(os.path.dirname(meta_path), exist_ok=True)
    with open(meta_path, "w", encoding="utf-8") as f:
        yaml.safe_dump(info.to_json(), f)


def chomp_extension(file_path: str) -> str:
    return Path(file_path).with_suffix("").name


def clean_filename(file_path: str) -> str:
    return Path(file_path).with_suffix("").name


def try_possible_extensions(file_path: str, base_path: str) -> str:
    for ext in ["", ".yaml", ".json", ".txt", ".md"]:
        possible = (Path(base_path) / file_path).with_suffix(f"{ext}")
        if possible.exists():
            return str(possible)
    raise ValueError(
        "Could not find file with any valid extensions (json, yaml, txt, md): "
        + file_path
    )


def read_md_file(file_path: str) -> dict | list[str]:
    with open(file_path) as f:
        data = frontmatter.load(f)
        return [line.strip("- *\t") for line in data.content.split("\n")]


def read_json_file(file_path: str) -> dict | list[str]:
    with open(file_path) as f:
        return json.load(f)


def read_yaml_file(file_path: str) -> dict | list[str]:
    with open(file_path) as f:
        return yaml.safe_load(f)


def read_txt_file(file_path: str) -> list[str]:
    with open(file_path) as f:
        return [line.strip() for line in f.readlines()]


def read_file(file_path: str, file_format: str) -> dict | list[str]:
    if file_format == "json":
        return read_json_file(file_path)
    elif file_format == "yaml":
        return read_yaml_file(file_path)
    elif file_format == "txt":
        return read_txt_file(file_path)
    elif file_format == "md":
        return read_md_file(file_path)
    elif file_format == "guess":
        if file_path.endswith(".json"):
            return read_json_file(file_path)
        elif file_path.endswith(".yaml") or file_path.endswith(".yml"):
            return read_yaml_file(file_path)
        elif file_path.endswith(".txt"):
            return read_txt_file(file_path)
        elif file_path.endswith(".md"):
            return read_md_file(file_path)
        else:
            raise ValueError("Could not guess file format")
    else:
        raise ValueError("Invalid file format")
